/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "gribouy-cursor.h"
#include "gribouy-triplet.h"
#include "gribouy-curve.h"

enum {
	PROP_0,
	PROP_TYPE,
	PROP_CHILDREN
};

static void gribouy_cursor_init (GribouyCursor *cursor);
static void gribouy_cursor_class_init (GribouyCursorClass *class);
static void gribouy_cursor_set_property (GObject *object, guint property_id,
		const GValue *value, GParamSpec *pspec);
static void gribouy_cursor_get_property (GObject *object, guint property_id,
		GValue *value, GParamSpec *pspec);
static void gribouy_cursor_select (GribouyObject *obj, gboolean v);
static void gribouy_cursor_edit (GribouyObject *obj, gboolean v);
static void gribouy_cursor_only_select (GribouyObject *obj);
static void gribouy_cursor_move_selection (GribouyObject *obj, gdouble x, gdouble y);
static void gribouy_cursor_open_menu (GribouyObject *obj, GdkEvent *event);
static void gribouy_cursor_erase_cb (GribouyCursor *cursor, guint action,
		GtkWidget *item);
static gint gribouy_cursor_event (GribouyCursor *cursor, GdkEvent *event, GnomeCanvasItem *center);
static void gribouy_cursor_update_path (GribouyCursor *cursor);
static void gribouy_cursor_destroy (GtkObject *object);
static void gribouy_cursor_update (GnomeCanvasItem *item, double *affine,
		ArtSVP *clip_path, int flags);

static GribouyObjectClass *parent_class = NULL;

GType
gribouy_cursor_get_type (void)
{
	static GType gribouy_cursor_type = 0;

	if (!gribouy_cursor_type) {
		static const GTypeInfo gribouy_cursor_info = {
			sizeof (GribouyCursorClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gribouy_cursor_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (GribouyCursor),
			0,
			(GInstanceInitFunc) gribouy_cursor_init
		};

		gribouy_cursor_type = g_type_register_static (
				GRIBOUY_TYPE_OBJECT, "GribouyCursor",
				&gribouy_cursor_info, 0);
	}
	return gribouy_cursor_type;
}

static void 
gribouy_cursor_class_init (GribouyCursorClass * class)
{
	GObjectClass *g_object_class = (GObjectClass *) class;
	GtkObjectClass *object_class = (GtkObjectClass *) class;
	GnomeCanvasItemClass *item_class = (GnomeCanvasItemClass *) class;
	GribouyObjectClass *obj_class = (GribouyObjectClass *) class;

	parent_class = g_type_class_peek_parent (class);

	g_object_class->set_property = gribouy_cursor_set_property;
	g_object_class->get_property = gribouy_cursor_get_property;

	g_object_class_install_property
		(g_object_class,
		 PROP_TYPE,
		 g_param_spec_uint ("type", NULL, NULL,
			 0, 2, 0,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));
	g_object_class_install_property
		(g_object_class,
		 PROP_CHILDREN,
		 g_param_spec_pointer ("children", NULL, NULL,
			 (G_PARAM_READABLE)));

	object_class->destroy = gribouy_cursor_destroy;
	obj_class->select = gribouy_cursor_select;
	obj_class->edit = gribouy_cursor_edit;
	obj_class->only_select = gribouy_cursor_only_select;
	obj_class->move_selection = gribouy_cursor_move_selection;
	obj_class->open_menu = gribouy_cursor_open_menu;
	item_class->update = gribouy_cursor_update;
}

static void 
gribouy_cursor_init (GribouyCursor *cursor)
{
	cursor->type = 0;
}

static void
gribouy_cursor_get_property (GObject *object, guint property_id,
		GValue *value, GParamSpec *pspec)
{
	GribouyCursor *cursor;

	cursor = GRIBOUY_CURSOR (object);

	switch (property_id) {
		case PROP_TYPE:
			g_value_set_uint (value, cursor->type);
			break;
		case PROP_CHILDREN:
			g_value_set_pointer (value, NULL);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
					property_id, pspec);
		break;
	}
}

static void
gribouy_cursor_set_property (GObject *object, guint param_id,
		const GValue *value, GParamSpec *pspec)
{
	GribouyObject *obj;
	GribouyCursor *cursor;

	obj = GRIBOUY_OBJECT (object);
	cursor = GRIBOUY_CURSOR (object);

	switch (param_id) {
		case PROP_TYPE:
			cursor->type = g_value_get_uint (value);
			if (!cursor->type)
				g_object_set (G_OBJECT (obj->center),
						"fill_color_rgba", 0xaaffff80,
						NULL);
			else
				g_object_set (G_OBJECT (obj->center),
						"fill_color_rgba", 0xaaff8080,
						NULL);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
					param_id, pspec);
			break;
	}
}

static void
gribouy_cursor_edit (GribouyObject *obj, gboolean v)
{
	GribouyCursor *cursor;

	cursor = GRIBOUY_CURSOR (obj);

	if (v)
		gnome_canvas_item_show ((GnomeCanvasItem *) obj);
	else
		gnome_canvas_item_hide ((GnomeCanvasItem *) obj);
}

static void
gribouy_cursor_select (GribouyObject *obj, gboolean v)
{
	GribouyCursor *cursor;

	cursor = GRIBOUY_CURSOR (obj);
	if (cursor->type == 0) {
		if (v)
			g_object_set (G_OBJECT (obj->center),
					"fill_color_rgba", 0xaaffffff,
					"outline_color", "red", NULL);
		else
			g_object_set (G_OBJECT (obj->center),
					"fill_color_rgba", 0xaaffff80,
					"outline_color", "blue", NULL);
	}
	else {
		if (v)
			gnome_canvas_item_show (GNOME_CANVAS_ITEM (obj));
		else
			gnome_canvas_item_hide (GNOME_CANVAS_ITEM (obj));
	}
}

static void
gribouy_cursor_only_select (GribouyObject *obj)
{
	GribouyTriplet *triplet;

	if (obj->selected)
		return;

	triplet = (GribouyTriplet *) GNOME_CANVAS_ITEM (obj)->parent;
	if ((GnomeCanvasItem *) obj == GRIBOUY_OBJECT (triplet)->center)
		return;
	else if (obj == triplet->ctrl1) {
		g_object_set (G_OBJECT (triplet->ctrl1), "selected", TRUE, NULL);
		if (triplet->ctrl2)
			g_object_set (G_OBJECT (triplet->ctrl2),
					"selected", FALSE, NULL);
	}
	else if (obj == triplet->ctrl2) {
		if (triplet->ctrl1)
			g_object_set (G_OBJECT (triplet->ctrl1),
					"selected", FALSE, NULL);
		g_object_set (G_OBJECT (triplet->ctrl2), "selected", TRUE, NULL);
	}
}


	
/* Creates a new cursor in the triplet at the position (x,y). If type = 0,
 * the cursor is a point else it is a control.*/
GribouyObject * 
gribouy_cursor_new (GribouyObject *triplet, double x, double y, guint type)
{
	GribouyCursor *cursor;
	GnomeCanvasItem *center;
	GnomeCanvasItem *rect;
	double pixels_per_unit;
	GnomeCanvasPoints *pts;
	double w;

	g_return_if_fail (GRIBOUY_IS_LETTER_WINDOW (triplet->window));
	pixels_per_unit = GNOME_CANVAS_ITEM (triplet)->canvas->pixels_per_unit;
	w = 3. / pixels_per_unit;

	cursor = GRIBOUY_CURSOR (gnome_canvas_item_new (
				(GnomeCanvasGroup *) triplet,
				gribouy_cursor_get_type (),
				"x", x,
				"y", y,
				"cursor", GDK_DRAFT_LARGE,
				NULL));

	if (x != 0 || y !=0) {
		pts = gnome_canvas_points_new (2);
		pts->coords[0] = 0.;
		pts->coords[1] = 0.;
		pts->coords[2] = -x;
		pts->coords[3] = -y;
		cursor->line = GNOME_CANVAS_LINE (gnome_canvas_item_new (
				(GnomeCanvasGroup *) cursor,
				gnome_canvas_line_get_type (),
				"points", pts,
				"width_pixels", 0,
				"fill_color", "blue",
				NULL));
		gnome_canvas_item_lower_to_bottom ((GnomeCanvasItem *)cursor->line);
	}
	else
		cursor->line = NULL;

	center = gnome_canvas_item_new ((GnomeCanvasGroup *) cursor,
			GNOME_TYPE_CANVAS_RECT,
			"x1", -w,
			"y1", -w,
			"x2", w,
			"y2", w,
			"width_pixels", 1,
			"outline_color", "blue",
			NULL);

	
	g_object_set (G_OBJECT (cursor),
			"center", center,
			"window", triplet->window,
			"type", type, NULL);
	g_return_if_fail (GRIBOUY_IS_LETTER_WINDOW (triplet->window));
	return (GribouyObject *) cursor;
}

static void
gribouy_cursor_update (GnomeCanvasItem *item, double *affine,
		ArtSVP *clip_path, int flags)
{
	GnomeCanvasRE *re;
	double w, h;
	double i2c[6], c2i[6];;

	g_object_get (G_OBJECT (item), "center", &re, NULL);

	gnome_canvas_item_i2c_affine ((GnomeCanvasItem *) item, i2c);
	art_affine_invert (c2i, i2c);

	w = c2i[0] * 3 + c2i[2] * 3;
	h = c2i[1] * 3 + c2i[3] * 3;
	if (re->x1 != -w || re->x2 != w || re->y1 != -h || re->y2 != h) {
		re->path_dirty = 1;
		re->x1 = -w;
		re->x2 = w;
		re->y1 = -h;
		re->y2 = h;
	}

	if (GNOME_CANVAS_ITEM_CLASS (parent_class)->update)
		(* GNOME_CANVAS_ITEM_CLASS (parent_class)->update) (item, affine, clip_path, flags);
}

static void
gribouy_cursor_destroy (GtkObject *object)
{
	g_return_if_fail (object != NULL);
	g_return_if_fail (GRIBOUY_IS_CURSOR (object));

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

gboolean inline
gribouy_is_cursor (GnomeCanvasItem *item)
{
	return (item && item->parent && GRIBOUY_IS_CURSOR (item->parent));
}

static void
gribouy_cursor_update_path (GribouyCursor *cursor)
{
	double cx, cy;
	GnomeCanvasPoints *pts;
	GribouyObject *triplet;

	triplet = GRIBOUY_OBJECT (GNOME_CANVAS_ITEM (cursor)->parent);
	if (cursor->line)
		gtk_object_destroy (GTK_OBJECT (cursor->line));

	g_object_get (G_OBJECT (cursor),
			"x", &cx, "y", &cy, NULL);
	if (cx != 0 || cy != 0) {
		pts = gnome_canvas_points_new (2);
		pts->coords[0] = 0.;
		pts->coords[1] = 0.;
		pts->coords[2] = -cx;
		pts->coords[3] = -cy;
		cursor->line = GNOME_CANVAS_LINE (gnome_canvas_item_new (
				(GnomeCanvasGroup *) cursor,
				gnome_canvas_line_get_type (),
				"points", pts,
				"width_pixels", 0,
				"fill_color", "blue",
				NULL));
		gnome_canvas_item_lower_to_bottom ((GnomeCanvasItem *)cursor->line);
		gnome_canvas_points_unref (pts);
	}
	else
		cursor->line = NULL;
}
	
static void
gribouy_cursor_move_selection (GribouyObject *cursor, gdouble x, gdouble y)
{
	GribouyTriplet *triplet;
	GribouyCursor *cur, *ctrl1, *ctrl2;
	double x1, y1, x2, y2;
	double l1, l2;

	triplet = GRIBOUY_TRIPLET (GNOME_CANVAS_ITEM (cursor)->parent);
	cur = GRIBOUY_CURSOR (cursor);
	if (((GribouyObject *)triplet)->center == (GnomeCanvasItem *)cursor)
		return;

	gnome_canvas_item_move (GNOME_CANVAS_ITEM (cursor),
			x, y);
	gribouy_cursor_update_path (cur);

	if (triplet->linked) {
		ctrl1 = GRIBOUY_CURSOR (triplet->ctrl1);
		ctrl2 = GRIBOUY_CURSOR (triplet->ctrl2);
		g_object_get (G_OBJECT (ctrl1),
				"x", &x1, "y", &y1, NULL);
		g_object_get (G_OBJECT (ctrl2),
				"x", &x2, "y", &y2, NULL);
		l1 = sqrt (x1 * x1 + y1 * y1);
		l2 = sqrt (x2 * x2 + y2 * y2);

		if (ctrl1 == cur) {
			if (l1 != 0) {
				x2 = -x1 * l2 / l1;
				y2 = -y1 * l2 / l1;
			}
			g_object_set (G_OBJECT (ctrl2),
					"x", x2, "y", y2, NULL);
			gnome_canvas_item_request_update (GRIBOUY_OBJECT (ctrl2)->center);
			gribouy_cursor_update_path (ctrl2);
		}
		else {
			if (l2 != 0) {
				x1 = -x2 * l1 / l2;
				y1 = -y2 * l1 / l2;
			}
			g_object_set (G_OBJECT (ctrl1),
					"x", x1, "y", y1, NULL);
			gnome_canvas_item_request_update (GRIBOUY_OBJECT (ctrl1)->center);
			gribouy_cursor_update_path (ctrl1);
		}
	}
}

static void
gribouy_cursor_open_menu (GribouyObject *obj, GdkEvent *event)
{
	GtkItemFactory *factory;
	GtkAccelGroup *accel;
	GtkWidget *menu;
	GtkItemFactoryEntry menu_items[] = {
		{ "/_Erase",
			NULL,
			gribouy_cursor_erase_cb, 1,
			"<StockItem>", GTK_STOCK_CLEAR },
	};

	if (GRIBOUY_CURSOR (obj)->type == 0)
		return;

	accel = gtk_accel_group_new ();
	factory = gtk_item_factory_new (GTK_TYPE_MENU,
			"<cursor>", accel);
	gtk_item_factory_create_items (factory, 1, menu_items, obj);
	menu = gtk_item_factory_get_widget (factory, "<cursor>");

	if (menu) {
		gtk_widget_show (menu);
		gtk_menu_popup (GTK_MENU (menu),
				NULL, NULL, NULL, NULL,
				2, event->button.time);
	}
}

static void
gribouy_cursor_erase_cb (GribouyCursor *cursor, guint action,
		GtkWidget *item)
{
	GList *lst;
	GribouyCurve *curve;
	GribouyTriplet *triplet, *triplet1;
	int index;

	triplet = GRIBOUY_TRIPLET (GNOME_CANVAS_ITEM (cursor)->parent);
	curve = GRIBOUY_CURVE (GNOME_CANVAS_ITEM (triplet)->parent);
	g_object_get (G_OBJECT (triplet), "index", &index, NULL);
	if (triplet->ctrl2 == (GribouyObject *) cursor) {
		triplet1 = (GribouyTriplet *) gribouy_curve_get_triplet_index (GRIBOUY_CURVE (curve), index + 1);
		gtk_object_destroy (GTK_OBJECT (cursor));
		triplet->ctrl2 = NULL;
		gtk_object_destroy (GTK_OBJECT (triplet1->ctrl1));
		triplet1->ctrl1 = NULL;
	}
	else if (triplet->ctrl1 == (GribouyObject *) cursor) {
		triplet1 = (GribouyTriplet *) gribouy_curve_get_triplet_index (GRIBOUY_CURVE (curve), index - 1);
		gtk_object_destroy (GTK_OBJECT (cursor));
		triplet->ctrl1 = NULL;
		gtk_object_destroy (GTK_OBJECT (triplet1->ctrl2));
		triplet1->ctrl2 = NULL;
	}
	g_signal_emit_by_name (G_OBJECT (triplet), "changed", 0);
}

