/* Gribouy : A type1 font editor
 * Copyright (C) 2003  David BOUCHER
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "gribouy-letter-window.h"
#include "gribouy-dialog.h"
#include "gribouy-grid.h"
#include "gribouy-letter.h"
#include "gribouy-marshal.h"
#include "gribouy-marshal.c"

static void gribouy_letter_window_finalize (GObject *object);
static void gribouy_letter_window_init (GribouyLetterWindow *window);
static void gribouy_letter_window_class_init (GribouyLetterWindowClass * class);
static void gribouy_letter_window_set_property (GObject *object,
		guint property_id, const GValue *value, GParamSpec *pspec);
static void gribouy_letter_window_get_property (GObject *object,
		guint property_id, GValue *value, GParamSpec *pspec);
static gint gribouy_letter_window_event (GribouyLetterWindow *window,
		GdkEvent *event, GnomeCanvas *canvas);
static void gribouy_letter_window_save (GribouyLetterWindow *window);
static void gribouy_letter_window_changed (GribouyLetterWindow *window, gboolean v);

static void on_canvas_tool_activate (GribouyLetterWindow *window,
		guint action, GtkWidget *widget);
static void on_canvas_zoom_activate (GribouyLetterWindow *window,
		guint action, GtkWidget *widget);
static void on_canvas_image_activate (GribouyLetterWindow *window,
		guint action, GtkWidget *widget);
static void gribouy_letter_window_grid_cb (GribouyLetterWindow *window,
		guint action, GtkWidget *widget);
static gboolean gribouy_letter_window_insert_image (GtkFileSelection *sel);
static void gribouy_letter_window_hruler_update (GtkRuler *ruler, GdkEventMotion      *event, GnomeCanvas *canvas);
static void gribouy_letter_window_vruler_update (GtkRuler *ruler, GdkEventMotion      *event, GnomeCanvas *canvas);
static gint gribouy_letter_window_scrollbar_update (GribouyLetterWindow *window, gpointer user_data);
static gint gribouy_letter_window_scroll_event (GtkWidget *widget, GdkEventScroll *event);
static void gribouy_letter_window_need_grid (GribouyLetterWindow *window,
		double x1, double y1, double x2, double y2);

enum {
	CHANGED_SIGNAL,
	NEED_GRID_SIGNAL,
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_LETTER_NAME
};

static guint letter_window_signals[LAST_SIGNAL] = { 0 };
static GtkWindowClass *parent_class = NULL;

GType
gribouy_letter_window_get_type (void)
{
	static GType gribouy_letter_window_type = 0;

	if (!gribouy_letter_window_type) {
		static const GTypeInfo gribouy_letter_window_info = {
			sizeof (GribouyLetterWindowClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gribouy_letter_window_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (GribouyLetterWindow),
			0,
			(GInstanceInitFunc) gribouy_letter_window_init,
		};

		gribouy_letter_window_type = g_type_register_static (
				GTK_TYPE_WINDOW,
				"GribouyLetterWindow", &gribouy_letter_window_info, 0);
	}
	return gribouy_letter_window_type;
}

static void
gribouy_letter_window_finalize (GObject *object)
{
	GribouyLetterWindow *window = GRIBOUY_LETTER_WINDOW (object);

	if (window->letter_name) {
		g_free (window->letter_name);
		window->letter_name = NULL;
	}
	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize)(object);
}

static void 
gribouy_letter_window_init (GribouyLetterWindow * window)
{
	window->grid = NULL;
	window->tool = TOOL_NONE;
	window->zoom = 0.5;
}

static void 
gribouy_letter_window_class_init (GribouyLetterWindowClass * class)
{
	GObjectClass *g_object_class = (GObjectClass *) class;
	GtkObjectClass *object_class = (GtkObjectClass *) class;
	GtkWidgetClass *widget_class = (GtkWidgetClass *) class;

	parent_class = g_type_class_ref (GTK_TYPE_WINDOW);

	g_object_class->finalize = gribouy_letter_window_finalize;
	g_object_class->get_property = gribouy_letter_window_get_property;
	g_object_class->set_property = gribouy_letter_window_set_property;

	letter_window_signals[CHANGED_SIGNAL] =
		g_signal_new ("changed",
				G_TYPE_FROM_CLASS (g_object_class),
				G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
				G_STRUCT_OFFSET (GribouyLetterWindowClass, changed),
				NULL, NULL,
				g_cclosure_marshal_VOID__BOOLEAN,
				G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
	letter_window_signals[NEED_GRID_SIGNAL] =
		g_signal_new ("need_grid",
				G_TYPE_FROM_CLASS (g_object_class),
				G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
				G_STRUCT_OFFSET (GribouyLetterWindowClass, need_grid),
				NULL, NULL,
				gribouy_marshal_VOID__DOUBLE_DOUBLE_DOUBLE_DOUBLE,
				G_TYPE_NONE, 4, G_TYPE_DOUBLE, G_TYPE_DOUBLE, G_TYPE_DOUBLE, G_TYPE_DOUBLE);

	g_object_class_install_property
		(g_object_class,
		 PROP_LETTER_NAME,
		 g_param_spec_string ("letter_name", NULL, NULL,
			 NULL,
			 (G_PARAM_READABLE | G_PARAM_WRITABLE)));

	class->changed = gribouy_letter_window_changed;
	class->need_grid = gribouy_letter_window_need_grid;
	widget_class->scroll_event = gribouy_letter_window_scroll_event;
}

static void
gribouy_letter_window_get_property (GObject *object, guint property_id,
		GValue *value, GParamSpec *pspec)
{
	GribouyLetterWindow *window = GRIBOUY_LETTER_WINDOW (object);

	switch (property_id) {
		case PROP_LETTER_NAME:
			g_value_set_string (value, window->letter_name);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
					property_id, pspec);
	}
}

static void
gribouy_letter_window_set_property (GObject *object, guint param_id,
		const GValue *value, GParamSpec *pspec)
{
	GribouyLetterWindow *window = GRIBOUY_LETTER_WINDOW (object);

	switch (param_id) {
		case PROP_LETTER_NAME:
			g_free (window->letter_name);
			window->letter_name = g_value_dup_string (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
					param_id, pspec);
	}
}

static gint
gribouy_letter_window_event (GribouyLetterWindow  *window, GdkEvent *event,
		GnomeCanvas *canvas)
{
	static double x1, y1;
	static gboolean dragging = FALSE;
	static GnomeCanvasRect *sel = NULL;
	int grid_x, grid_y;
	ArtPoint point;
	GnomeCanvasItem *item;
	GtkWidget *canvas_menu, *menu;
	GtkItemFactory *factory;
	gboolean magnetic;
	GtkAccelGroup *accel;

	GtkItemFactoryEntry menu_items[] = {
		{ "/_Edit",
			"<control>E",
			on_canvas_tool_activate, 1,
			"<StockItem>", GTK_STOCK_JUMP_TO },
		{ "/Multi-_Line",
			"<control>L",
			on_canvas_tool_activate, 2,
			"<StockItem>", GTK_STOCK_ADD },
		{ "/_Insert an image",
			"<control>I",
			on_canvas_image_activate, 1 },
		{ "/_Zoom in",
			"<control>+",
			on_canvas_zoom_activate, 1,
			"<StockItem>", GTK_STOCK_ZOOM_IN },
		{ "/Zoom _out",
			"<control>-",
			on_canvas_zoom_activate, 2,
			"<StockItem>", GTK_STOCK_ZOOM_OUT },
		{ "/_Grid",
			"<control>G",
			gribouy_letter_window_grid_cb, 1 },
	};

	g_object_get (G_OBJECT (window->grid),
			"grid_x", &grid_x,
			"grid_y", &grid_y,
			"magnetic", &magnetic,
			NULL);

	point.x = event->button.x;
	point.y = event->button.y;
	gnome_canvas_c2w (canvas, point.x, point.y,
			&point.x, &point.y);

	item = gnome_canvas_get_item_at (canvas, point.x, point.y);

	switch (event->type) {
	case GDK_BUTTON_PRESS:
		switch (event->button.button) {
		case 1:
			x1 = point.x;
			y1 = point.y;
			if ((!item || !GRIBOUY_IS_OBJECT (item->parent)) && window->tool == TOOL_NONE)
				dragging = TRUE;
			break;
		case 3:
			/* Initialization of menus */
			accel = gtk_accel_group_new ();
			factory = gtk_item_factory_new (GTK_TYPE_MENU,
					"<canvas>", accel);
			gtk_item_factory_create_items (factory, 6,
					menu_items, window);

			canvas_menu = gtk_item_factory_get_widget (
					factory, "<canvas>");
			if (canvas_menu) {
				gtk_widget_show (canvas_menu);
				window->pt = point;
				gtk_menu_popup (GTK_MENU (canvas_menu),
						NULL, NULL, NULL, window,
						3, event->button.time);
			}
			break;
		}
		break;
	case GDK_BUTTON_RELEASE:
		if (event->button.button == 1) {
			if (dragging) {
				gribouy_object_select_children (GRIBOUY_OBJECT (window->letter),
						sel,
						event->button.state & GDK_CONTROL_MASK,
						TRUE);
				if (sel) {
					gtk_object_destroy (GTK_OBJECT (sel));
					sel = NULL;
				}
				dragging = FALSE;
			}
			else if (window->tool == TOOL_LINE) {
				if (gribouy_is_cursor (item)) {
					g_object_get (G_OBJECT (item->parent),
							"x", &point.x,
							"y", &point.y,
							NULL);
					gnome_canvas_item_i2w (item->parent, &point.x, &point.y);
				}
				else if (magnetic) {
					point.x = ROUND (point.x / grid_x) * grid_x;
					point.y = ROUND (point.y / grid_y) * grid_y;
				}
				gnome_canvas_item_w2i (window->letter,
						&point.x, &point.y);
				gribouy_letter_append_point (GRIBOUY_OBJECT (window->letter),
						&point);
			}
		}
		break;
	case GDK_MOTION_NOTIFY:
		if (dragging) {
			if (!sel) {
				sel = GNOME_CANVAS_RECT (gnome_canvas_item_new (
							gnome_canvas_root (GNOME_CANVAS (canvas)),
							gnome_canvas_rect_get_type (),
							"x1", x1,
							"y1", y1,
							"x2", point.x,
							"y2", point.y,
							"fill_color_rgba", 0xeecccc80,
							"outline_color", "red",
							"width_pixels", 0,
							NULL));
			}
			else
				g_object_set (G_OBJECT (sel),
						"x2", point.x,
						"y2", point.y, NULL);
		}
	default:
		break;
	}
	return FALSE;
}

GtkWidget *
gribouy_letter_window_new (char *title, GribouyType1 *font)
{
	GribouyLetterWindow *window;
	GnomeCanvasGroup *group;
	GtkTable *table;
	guint width;
	gint left;
	GnomeCanvas *canvas;
	GtkWidget *widget;
	GList *prog;
	GArray *blue_values;

	window = GRIBOUY_LETTER_WINDOW (g_object_new (
				gribouy_letter_window_get_type (), NULL));
	GTK_WINDOW (window)->type = GTK_WINDOW_TOPLEVEL;
	gtk_widget_set_usize (GTK_WIDGET (window), 300, 400);
	gtk_window_set_title (GTK_WINDOW (window), title);

	canvas = GNOME_CANVAS (gnome_canvas_new_aa ());
	window->canvas = GTK_WIDGET (canvas);
	g_signal_connect_swapped (G_OBJECT (canvas), "event",
			G_CALLBACK (gribouy_letter_window_event),
			window);
	g_signal_connect (G_OBJECT (window), "destroy",
			G_CALLBACK (gribouy_letter_window_save), NULL);
	gnome_canvas_set_pixels_per_unit (canvas, window->zoom);
	font->current = g_strdup (title);
	window->letter_name = g_strdup (title);
	g_object_get (G_OBJECT (font), "symbol", &prog, NULL);
	window->font = font;
	group = gnome_canvas_root (canvas);
	window->letter = GNOME_CANVAS_ITEM (gribouy_letter_new_with_art (window, prog, font));

	g_object_get (G_OBJECT (font),
			"blue_values", &blue_values, NULL);
	g_object_get (G_OBJECT(window->letter),
			"left", &left,
			"width", &width,
			NULL);

	g_print ("width=%d\n", width);
	window->grid = gribouy_grid_new (window, blue_values, left, width);

	table = GTK_TABLE (gtk_table_new (3, 3, FALSE));
	gtk_container_add (GTK_CONTAINER (window), GTK_WIDGET (table));

	window->h_ruler = gtk_hruler_new ();
	g_signal_connect_swapped (G_OBJECT (canvas),
			"motion_notify_event",
			G_CALLBACK (gribouy_letter_window_hruler_update),
			G_OBJECT (window->h_ruler));

	window->v_ruler = gtk_vruler_new ();
	g_signal_connect_swapped (G_OBJECT (canvas),
			"motion_notify_event",
			G_CALLBACK (gribouy_letter_window_vruler_update),
			G_OBJECT (window->v_ruler));

	canvas->layout.hadjustment->step_increment = 5;
	canvas->layout.vadjustment->step_increment = 5;
	window->h_scrollbar = gtk_hscrollbar_new (canvas->layout.hadjustment);
	g_signal_connect_swapped (G_OBJECT (window->h_scrollbar),
			"value_changed",
			G_CALLBACK (gribouy_letter_window_scrollbar_update),
			window);

	window->v_scrollbar = gtk_vscrollbar_new (canvas->layout.vadjustment);
	g_signal_connect_swapped (G_OBJECT (window->v_scrollbar),
			"value_changed",
			G_CALLBACK (gribouy_letter_window_scrollbar_update),
			window);
	g_signal_connect (G_OBJECT (window),
			"expose_event",
			G_CALLBACK (gribouy_letter_window_scrollbar_update),
			NULL);

	gtk_table_attach (table, window->h_ruler,
			1, 2, 0, 1,
			GTK_EXPAND | GTK_SHRINK | GTK_FILL, GTK_FILL,
			0, 0);
	gtk_table_attach (table, window->v_ruler,
			0, 1, 1, 2,
			GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL,
			0, 0);
	gtk_table_attach (table, window->canvas,
			1, 2, 1, 2,
			GTK_EXPAND | GTK_SHRINK | GTK_FILL,
			GTK_EXPAND | GTK_SHRINK | GTK_FILL,
			0, 0);
	gtk_table_attach (table, window->h_scrollbar,
			0, 2, 2, 3,
			GTK_EXPAND | GTK_SHRINK | GTK_FILL, GTK_FILL,
			0, 0);
	gtk_table_attach (table, window->v_scrollbar,
			2, 3, 0, 2,
			GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL,
			0, 0);
	return (GtkWidget *)window;
}

static void
on_canvas_tool_activate (GribouyLetterWindow *window,
		guint action, GtkWidget *widget)
{
	switch (action) {
		case 1:
			g_object_set (G_OBJECT (window->letter),
					"editable", TRUE,
					NULL);
			window->tool = TOOL_NONE;
			break;
		case 2:
			g_object_set (G_OBJECT (window->letter),
					"editable", FALSE, NULL);
			window->tool = TOOL_LINE;
			break;
	}
}

static void
on_canvas_zoom_activate (GribouyLetterWindow *window,
		guint action, GtkWidget *widget)
{
	if (action == 1)
		window->zoom *= 1.3;
	else
		window->zoom /= 1.3;

	gnome_canvas_set_pixels_per_unit (GNOME_CANVAS_ITEM (window->letter)->canvas, window->zoom);
	gribouy_grid_construct (GRIBOUY_GRID (window->grid));
}

static void
on_canvas_image_activate (GribouyLetterWindow *window,
		guint action, GtkWidget *widget)
{
	GtkFileSelection *file_dlg;

	file_dlg = GTK_FILE_SELECTION (gtk_file_selection_new ("Insert an image..."));
	gtk_signal_connect (GTK_OBJECT (file_dlg), "delete_event",
			(GtkSignalFunc)gtk_widget_destroy, NULL);
	gtk_signal_connect (GTK_OBJECT (file_dlg), "destroy",
			(GtkSignalFunc)gtk_widget_destroy, NULL);

	gtk_signal_connect_object (GTK_OBJECT (file_dlg->cancel_button),
				   "clicked",
				   (GtkSignalFunc) gtk_widget_destroy,
				   GTK_OBJECT (file_dlg));

	gtk_signal_connect_object (GTK_OBJECT (file_dlg->ok_button),
				   "clicked",
				   (GtkSignalFunc) gribouy_letter_window_insert_image,
				   GTK_OBJECT (file_dlg));

	gtk_object_set_data (GTK_OBJECT (file_dlg), "window", window);
	gtk_widget_show (GTK_WIDGET (file_dlg));
}

static void
gribouy_letter_window_grid_cb (GribouyLetterWindow *window,
		guint action, GtkWidget *widget)
{
	GtkWidget *dialog;

	dialog = gribouy_grid_dialog_new (window);
}


static gboolean
gribouy_letter_window_insert_image (GtkFileSelection *sel)
{
	gchar *file;
	gchar *error_str;
	GribouyLetterWindow *window;

	window = GRIBOUY_LETTER_WINDOW (gtk_object_get_data (GTK_OBJECT (sel),
				"window"));
	file = (gchar *)gtk_file_selection_get_filename (sel);
	gtk_widget_destroy (GTK_WIDGET (sel));

	gribouy_bg_new (window, file,
				window->pt.x, window->pt.y);
}

static void
gribouy_letter_window_save (GribouyLetterWindow *window)
{
	GList *prog;

	g_print ("Saving %s\n", window->letter_name);
	if (strcmp (window->font->current, window->letter_name)) {
		if (window->font->current)
			g_free (window->font->current);
		window->font->current = g_strdup (window->letter_name);
	}
	prog = gribouy_letter_save (GRIBOUY_LETTER (window->letter));
	if (prog) {
		g_object_set (G_OBJECT (window->font),
				"symbol", prog, NULL);
		g_signal_emit (G_OBJECT (window),
				letter_window_signals[CHANGED_SIGNAL], 0, TRUE);
	}
	else
		g_signal_emit (G_OBJECT (window),
				letter_window_signals[CHANGED_SIGNAL], 0, FALSE);
	gtk_widget_destroy (GTK_WIDGET (window));
}

static void
gribouy_letter_window_hruler_update (GtkRuler *ruler, GdkEventMotion *event,
		GnomeCanvas *canvas)
{
	GdkEventMotion new_event;

	memcpy (&new_event, event, sizeof (GdkEventMotion));

	new_event.x -= canvas->layout.hadjustment->value;
	GTK_WIDGET_GET_CLASS (ruler)->motion_notify_event (GTK_WIDGET (ruler), &new_event);
}

static void
gribouy_letter_window_vruler_update (GtkRuler *ruler, GdkEventMotion *event,
		GnomeCanvas *canvas)
{
	GdkEventMotion new_event;

	memcpy (&new_event, event, sizeof (GdkEventMotion));

	new_event.y -= canvas->layout.vadjustment->value;
	GTK_WIDGET_GET_CLASS (GTK_WIDGET (ruler))->motion_notify_event (GTK_WIDGET (ruler), &new_event);
}

static gint
gribouy_letter_window_scrollbar_update (GribouyLetterWindow *window, gpointer user_data)
{
	GnomeCanvas *canvas;
	double x1, y1, x2, y2;
	double cx1, cy1, cx2, cy2;
	int canvas_height, canvas_width;

	canvas = GNOME_CANVAS (window->canvas);
	canvas_height = GTK_WIDGET (canvas)->allocation.height;
	canvas_width = GTK_WIDGET (canvas)->allocation.width;
	x1 = canvas->layout.hadjustment->value - canvas->zoom_xofs;
	y1 = canvas->layout.vadjustment->value - canvas->zoom_yofs;
	x2 = x1 + canvas_width;
	y2 = y1 + canvas_height;
	gnome_canvas_window_to_world (canvas, x1, y1, &x1, &y1);
	gnome_canvas_window_to_world (canvas, x2, y2, &x2, &y2);
/*
	gnome_canvas_c2w (canvas, x1, y1, &x1, &y1);
	gnome_canvas_c2w (canvas, x2, y2, &x2, &y2);
*/
	gnome_canvas_get_scroll_region (canvas, &cx1, &cy1, &cx2, &cy2);
	if (x1 < cx1)	cx1 = x1;
	if (x2 > cx2)	cx2 = x2;
	if (y1 < cy1)	cy1 = y1;
	if (y2 > cy2)	cy2 = y2;
	gnome_canvas_set_scroll_region (canvas, cx1, cy1, cx2, cy2);

	gtk_ruler_set_range (GTK_RULER (window->h_ruler),
			x1, x2,
			(x1 + x2) / 2,
			x2);
	gtk_ruler_set_range (GTK_RULER (window->v_ruler),
			-y1, -y2,
			(-y1 - y2) / 2,
			-y2);

	g_signal_emit (G_OBJECT (window),
			letter_window_signals[NEED_GRID_SIGNAL], 0, x1, y1, x2, y2);

	return FALSE;
}

static void
gribouy_letter_window_need_grid (GribouyLetterWindow *window, double x1, double y1, double x2, double y2)
{
}

static void
gribouy_letter_window_changed (GribouyLetterWindow *window, gboolean v)
{
}

static gint
gribouy_letter_window_scroll_event (GtkWidget *widget, GdkEventScroll *event)
{
	GribouyLetterWindow *window = (GribouyLetterWindow *)widget;
	GtkAdjustment *adj = GTK_RANGE (window->v_scrollbar)->adjustment;
	double pos;

	if (event->direction == GDK_SCROLL_DOWN)
		pos = adj->value + adj->page_increment / 4;
	else if (event->direction == GDK_SCROLL_UP)
		pos = adj->value - adj->page_increment / 4;

	gtk_adjustment_set_value (adj, pos);

	return TRUE;
}
